/* Copyright (c) 1997-1999 Miller Puckette and Ted Apel.
* For information on usage and redistribution, and for a DISCLAIMER OF ALL
* WARRANTIES, see the file, "LICENSE.txt," in this distribution.  */

/*
 * Fiddle is a pitch tracker hardwired to have hop size ("H") equal to
 * half its window size ("N").
 *
 * This version should compile for Max "0.26," JMAX, Pd, or Max/MSP.
 *
 * The "lastanalysis" field holds the shifted FT of the previous H
 * samples.  The buffer contains in effect points 1/2,  3/2, ..., (N-1)/2
 * of the DTFT of a real vector of length N, half of whose points are zero,
 * i.e.,  only the first H points are used.  Put another way, we get the
 * the odd-numbered points of the FFT of the H points, zero padded to 4*H in
 * length. The integer points 0, 1, ..., H-1
 * are found by interpolating these others,  using the fact that the
 * half-integer points are band-limited (they only have positive frequencies.)
 * To facilitate the interpolation the "lastanalysis" buffer contains
 * FILTSIZE extra points (1/2-FILTSIZE, ...,  -1/2) at the beginning and
 * FILTSIZE again at the end ((N+1)/2, ..., FILTSIZE+(N-1)/2).  The buffer
 * therefore has N+4*FILTSIZE floating-point numbers in it.
 *
 * after doing this I found out that you can just do a real FFT
 * of the H new points, zero-padded to contain N points, and using a similar
 * but simpler interpolation scheme you can still get 2N points of the DTFT
 * of the N points.  Jean Laroche is a big fat hen.
 *
 */

#define flog log
#define fexp exp
#define fsqrt sqrt

#include <cstdio>
#include <cstdarg>
static void post(const char*fmt, ...) { va_list ap;va_start(ap,fmt);vfprintf(stderr,fmt,ap);va_end(ap); fprintf(stderr,"\n"); }
static void error(const char*fmt, ...) { va_list ap;va_start(ap,fmt);vfprintf(stderr,fmt,ap);va_end(ap); fprintf(stderr,"\n"); }

#include <cmath>

#include "fft-simple.cc"


#define MINBIN 3
#define DEFAMPLO 40
#define DEFAMPHI 50
#define DEFATTACKTIME 100
#define DEFATTACKTHRESH 10
#define DEFVIBTIME 50
#define DEFVIBDEPTH 0.5
#define GLISS 0.7f
#define DBFUDGE 30.8f
#define MINFREQINBINS 5     /* minimum frequency in bins for reliable output */

#define MAXNPITCH 3
#define MAXHIST 3                   /* find N hottest peaks in histogram */

#define MAXPOINTS 8192
#define MINPOINTS 128
#define DEFAULTPOINTS 1024

#define HISTORY 20
#define MAXPEAK 100             /* maximum number of peaks */
#define DEFNPEAK 20             /* default number of peaks */

#define MAXNPEAK (MAXLOWPEAK + MAXSTRONGPEAK)
#define MINBW (0.03f)                   /* consider BW >= 0.03 FFT bins */

#define BINPEROCT 48                    /* bins per octave */
#define BPERO_OVER_LOG2 69.24936196f    /* BINSPEROCT/log(2) */
#define FACTORTOBINS (double)(4/0.0145453)       /* 4 / (pow(2.,1/48.) - 1) */
#define BINGUARD 10                     /* extra bins to throw in front */
#define PARTIALDEVIANCE 0.023f          /* acceptable partial detuning in % */
#define LOGTODB 4.34294481903f          /* 20/log(10) */

#define KNOCKTHRESH 10.f     /* don't know how to describe this */


static const double sigfiddle_partialonset[] =
{
0,
48,
76.0782000346154967102,
96,
111.45254855459339269887,
124.07820003461549671089,
134.75303625876499715823,
144,
152.15640006923099342109,
159.45254855459339269887,
166.05271769459026829915,
172.07820003461549671088,
177.62110647077242370064,
182.75303625876499715892,
187.53074858920888940907,
192,
};

#define NPARTIALONSET ((int)(sizeof(sigfiddle_partialonset)/sizeof(double)))

static const int sigfiddle_intpartialonset[] =
{
0,
48,
76,
96,
111,
124,
135,
144,
152,
159,
166,
172,
178,
183,
188,
192,
};

/* these coefficients, which come from the "upsamp" subdirectory,
are a filter kernel for upsampling by a factor of two, assuming
the sound to be upsampled has no energy above half the Nyquist, i.e.,
that it's already 2x oversampled compared to the theoretically possible
sample rate.  I got these by trial and error. */

#define FILT1 ((double)(.5 * 1.227054))
#define FILT2 ((double)(.5 * -0.302385))
#define FILT3 ((double)(.5 * 0.095326))
#define FILT4 ((double)(.5 * -0.022748))
#define FILT5 ((double)(.5 * 0.002533))
#define FILTSIZE 5

typedef struct peakout      /* a peak for output */
{
    double po_freq;                  /* frequency in hz */
    double po_amp;                   /* amplitude */
} t_peakout;

typedef struct peak         /* a peak for analysis */
{
    double p_freq;                   /* frequency in bins */
    double p_width;                  /* peak width in bins */
    double p_pow;                    /* peak power */
    double p_loudness;               /* 4th root of power */
    double *p_fp;                    /* pointer back to spectrum */
} t_peak;

typedef struct histopeak
{
    double h_pitch;                  /* estimated pitch */
    double h_value;                  /* value of peak */
    double h_loud;                   /* combined strength of found partials */
    int h_index;                    /* index of bin holding peak */
    int h_used;                     /* true if an x_hist entry points here */
} t_histopeak;

typedef struct pitchhist            /* struct for keeping history by pitch */
{
    double h_pitch;                  /* pitch to output */
    double h_amps[HISTORY];          /* past amplitudes */
    double h_pitches[HISTORY];       /* past pitches */
    double h_noted;                  /* last pitch output */
    int h_age;                      /* number of frames pitch has been there */
    t_histopeak *h_wherefrom;       /* new histogram peak to incorporate */
} t_pitchhist;

typedef struct sigfiddle                    /* instance struct */
{
    double *x_inbuf;                 /* buffer to analyze, npoints/2 elems */
    double *x_lastanalysis;          /* FT of last buffer (see main comment) */
    double *x_spiral;                /* 1/4-wave complex exponential */
    t_peakout *x_peakbuf;           /* spectral peaks for output */
    int x_npeakout;                 /* number of spectral peaks to output */
    int x_npeakanal;                /* number of spectral peaks to analyze */
    int x_phase;                    /* number of points since last output */
    int x_histphase;                /* phase into amplitude history vector */
    int x_hop;                      /* period of output, npoints/2 */
    double x_sr;                     /* sample rate */
    t_pitchhist x_hist[MAXNPITCH];  /* history of current pitches */
    int x_nprint;                   /* how many periods to print */
    int x_npitch;                   /* number of simultaneous pitches */
    double x_dbs[HISTORY];           /* DB history, indexed by "histphase" */
    double x_peaked;                 /* peak since last attack */
    int x_dbage;                    /* number of bins DB has met threshold */
    int x_auto;                     /* true if generating continuous output */
/* parameters */
    double x_amplo;
    double x_amphi;
    int x_attacktime;
    int x_attackbins;
    double x_attackthresh;
    int x_vibtime;
    int x_vibbins;
    double x_vibdepth;
    double x_npartial;
/* outlets & clock */
    int x_attackvalue;
} t_sigfiddle;

#if CHECKER
double fiddle_checker[1024];
#endif

static int sigfiddle_ilog2(int n)
{
    int ret = -1;
    while (n)
    {
        n >>= 1;
        ret++;
    }
    return (ret);
}

static double fiddle_mtof(double f)
{
    return (8.17579891564 * exp(.0577622650 * f));
}

static double fiddle_ftom(double f)
{
    return (17.3123405046 * log(.12231220585 * f));
}
#define ftom fiddle_ftom
#define mtof fiddle_mtof

static void sigfiddle_doit(t_sigfiddle *x)
{
    double spect1[4*MAXPOINTS];
    double spect2[MAXPOINTS + 4*FILTSIZE];
#if CHECKER
    double checker3[4*MAXPOINTS];
#endif

    t_peak peaklist[MAXPEAK + 1], *pk1;
    t_peakout *pk2;
    t_histopeak histvec[MAXHIST], *hp1;
    int i, j, k, hop = x->x_hop, n = 2*hop, npeak, npitch,
        logn = sigfiddle_ilog2(n), newphase, oldphase;
    double *fp, total_power, total_loudness, total_db;
    double maxbin = BINPEROCT * (logn-2),  *histogram = spect2 + BINGUARD;
    t_pitchhist *phist;
    double hzperbin = x->x_sr / (2.0f * n);
    int npeakout = x->x_npeakout, npeakanal = x->x_npeakanal;
    int npeaktot = (npeakout > npeakanal ? npeakout : npeakanal);

    oldphase = x->x_histphase;
    newphase = x->x_histphase + 1;
    if (newphase == HISTORY) newphase = 0;
    x->x_histphase = newphase;

        /*
         * multiply the H points by a 1/4-wave complex exponential,
         * and take FFT of the result.
         */
    { double*fp1, *fp2, *fp3;
    for (i = 0, fp1 = x->x_inbuf, fp2 = x->x_spiral, fp3 = spect1;
        i < hop; i++, fp1++, fp2 += 2, fp3 += 2)
            fp3[0] = fp1[0] * fp2[0], fp3[1] = fp1[0] * fp2[1]; }

    pd_fft(spect1, hop, 0); /* pd_fft */
    
        /*
         * now redistribute the points to get in effect the odd-numbered
         * points of the FFT of the H points, zero padded to 4*H in length.
         */
    { double*fp1, *fp2;
    for (i = 0, fp1 = spect1, fp2 = spect2 + (2*FILTSIZE);
        i < (hop>>1); i++, fp1 += 2, fp2 += 4)
            fp2[0] = fp1[0], fp2[1] = fp1[1];
    for (i = 0, fp1 = spect1 + n - 2, fp2 = spect2 + (2*FILTSIZE+2);
        i < (hop>>1); i++, fp1 -= 2, fp2 += 4)
            fp2[0] = fp1[0], fp2[1] = -fp1[1];
    for (i = 0, fp1 = spect2 + (2*FILTSIZE), fp2 = spect2 + (2*FILTSIZE-2);
        i<FILTSIZE; i++, fp1+=2, fp2-=2)
            fp2[0] = fp1[0],  fp2[1] = -fp1[1];
    for (i = 0, fp1 = spect2 + (2*FILTSIZE+n-2), fp2 = spect2 + (2*FILTSIZE+n);
        i<FILTSIZE; i++, fp1-=2, fp2+=2)
            fp2[0] = fp1[0],  fp2[1] = -fp1[1];
    }
#if 0
    {
        fp = spect2 + 2*FILTSIZE;
        post("x1 re %12.4f %12.4f %12.4f %12.4f %12.4f",
            fp[0], fp[2], fp[4], fp[6], fp[8]);
        post("x1 im %12.4f %12.4f %12.4f %12.4f %12.4f",
            fp[1], fp[3], fp[5], fp[7], fp[9]);
    }
#endif
        /* spect2 is now prepared; now combine spect2 and lastanalysis into
         * spect1.  Odd-numbered points of spect1 are the points of "last"
         * plus (-i, i, -i, ...) times spect1.  Even-numbered points are
         * the interpolated points of "last" plus (1, -1, 1, ...) times the
         * interpolated points of spect1.
         *
         * To interpolate, take FILT1 exp(-pi/4) times
         * the previous point,  FILT2*exp(-3*pi/4) times 3 bins before,
         * etc,  and FILT1 exp(pi/4), FILT2 exp(3pi/4), etc., to weight
         * the +1, +3, etc., points.
         *
         * In this calculation,  we take (1, i, -1, -i, 1) times the
         * -9, -7, ..., -1 points, and (i, -1, -i, 1, i) times the 1, 3,..., 9
         * points of the OLD spectrum, alternately adding and subtracting
         * the new spectrum to the old; then we multiply the whole thing
         * by exp(-i pi/4).
         */
    {double*fp1, *fp2, *fp3;
    for (i = 0, fp1 = spect1, fp2 = x->x_lastanalysis + 2*FILTSIZE,
        fp3 = spect2 + 2*FILTSIZE;
            i < (hop>>1); i++)
    {
        double re,  im;

        re= FILT1 * ( fp2[ -2] -fp2[ 1]  +fp3[ -2] -fp3[ 1]) +
            FILT2 * ( fp2[ -3] -fp2[ 2]  +fp3[ -3] -fp3[ 2]) +
            FILT3 * (-fp2[ -6] +fp2[ 5]  -fp3[ -6] +fp3[ 5]) +
            FILT4 * (-fp2[ -7] +fp2[ 6]  -fp3[ -7] +fp3[ 6]) +
            FILT5 * ( fp2[-10] -fp2[ 9]  +fp3[-10] -fp3[ 9]);

        im= FILT1 * ( fp2[ -1] +fp2[ 0]  +fp3[ -1] +fp3[ 0]) +
            FILT2 * (-fp2[ -4] -fp2[ 3]  -fp3[ -4] -fp3[ 3]) +
            FILT3 * (-fp2[ -5] -fp2[ 4]  -fp3[ -5] -fp3[ 4]) +
            FILT4 * ( fp2[ -8] +fp2[ 7]  +fp3[ -8] +fp3[ 7]) +
            FILT5 * ( fp2[ -9] +fp2[ 8]  +fp3[ -9] +fp3[ 8]);

        fp1[0] = 0.7071f * (re + im);
        fp1[1] = 0.7071f * (im - re);
        fp1[4] = fp2[0] + fp3[1];
        fp1[5] = fp2[1] - fp3[0];
        
        fp1 += 8, fp2 += 2, fp3 += 2;
        re= FILT1 * ( fp2[ -2] -fp2[ 1]  -fp3[ -2] +fp3[ 1]) +
            FILT2 * ( fp2[ -3] -fp2[ 2]  -fp3[ -3] +fp3[ 2]) +
            FILT3 * (-fp2[ -6] +fp2[ 5]  +fp3[ -6] -fp3[ 5]) +
            FILT4 * (-fp2[ -7] +fp2[ 6]  +fp3[ -7] -fp3[ 6]) +
            FILT5 * ( fp2[-10] -fp2[ 9]  -fp3[-10] +fp3[ 9]);

        im= FILT1 * ( fp2[ -1] +fp2[ 0]  -fp3[ -1] -fp3[ 0]) +
            FILT2 * (-fp2[ -4] -fp2[ 3]  +fp3[ -4] +fp3[ 3]) +
            FILT3 * (-fp2[ -5] -fp2[ 4]  +fp3[ -5] +fp3[ 4]) +
            FILT4 * ( fp2[ -8] +fp2[ 7]  -fp3[ -8] -fp3[ 7]) +
            FILT5 * ( fp2[ -9] +fp2[ 8]  -fp3[ -9] -fp3[ 8]);

        fp1[0] = 0.7071f * (re + im);
        fp1[1] = 0.7071f * (im - re);
        fp1[4] = fp2[0] - fp3[1];
        fp1[5] = fp2[1] + fp3[0];
        
        fp1 += 8, fp2 += 2, fp3 += 2;
    }
    }
#if 0
    if (x->x_nprint)
    {
        for (i = 0,  fp = spect1; i < 16; i++,  fp+= 4)
            post("spect %d %f %f --> %f", i, fp[0], fp[1],
                sqrt(fp[0] * fp[0] + fp[1] * fp[1]));
    }
#endif
         /* copy new spectrum out */
    {double*fp1,*fp2;
    for (i = 0, fp1 = spect2, fp2 = x->x_lastanalysis;
            i < n + 4*FILTSIZE; i++) *fp2++ = *fp1++;
    }

    for (i = 0; i < MINBIN; i++) spect1[4*i + 2] = spect1[4*i + 3] = 0;
        /* starting at bin MINBIN, compute hanning windowed power spectrum */
    {double*fp1;
    for (i = MINBIN, fp1 = spect1+4*MINBIN, total_power = 0;
        i < n-2; i++,  fp1 += 4)
    {
        double re = fp1[0] - 0.5f * (fp1[-8] + fp1[8]);
        double im = fp1[1] - 0.5f * (fp1[-7] + fp1[9]);
        fp1[3] = (total_power += (fp1[2] = re * re + im * im));
    }}

    if (total_power > 1e-9f)
    {
        total_db = (100.f - DBFUDGE) + LOGTODB * log(total_power/n);
        total_loudness = fsqrt(fsqrt(total_power));
        if (total_db < 0) total_db = 0;
    }
    else total_db = total_loudness = 0;
        /*  store new db in history vector */
    x->x_dbs[newphase] = total_db;
    if (total_db < x->x_amplo) goto nopow;
#if 1
    if (x->x_nprint) post("power %f", total_power);
#endif

#if CHECKER
        /* verify that our FFT resampling thing is putting out good results */
    for (i = 0; i < hop; i++)
    {
        checker3[2*i] = fiddle_checker[i];
        checker3[2*i + 1] = 0;
        checker3[n + 2*i] = fiddle_checker[i] = x->x_inbuf[i];
        checker3[n + 2*i + 1] = 0;
    }
    for (i = 2*n; i < 4*n; i++) checker3[i] = 0;
    fft(checker3, 2*n, 0);
    if (x->x_nprint)
    {
        for (i = 0,  fp = checker3; i < 16; i++,  fp += 2)
            post("spect %d %f %f --> %f", i, fp[0], fp[1],
                sqrt(fp[0] * fp[0] + fp[1] * fp[1]));
    }

#endif
    npeak = 0;

         /* search for peaks */
    for (i = MINBIN, fp = spect1+4*MINBIN, pk1 = peaklist;
        i < n-2 && npeak < npeaktot; i++, fp += 4)
    {
        double height = fp[2], h1 = fp[-2], h2 = fp[6];
        double totalfreq, pfreq, f1, f2, m, var, stdev;
        
        if (height < h1 || height < h2 ||
            h1 < 0.00001f*total_power || h2 < 0.00001f*total_power)
                continue;

            /* use an informal phase vocoder to estimate the frequency.
            Do this for the two adjacent bins too. */
        pfreq= ((fp[-8] - fp[8]) * (2.0f * fp[0] - fp[8] - fp[-8]) +
                (fp[-7] - fp[9]) * (2.0f * fp[1] - fp[9] - fp[-7])) /
                    (2.0f * height);
        f1=    ((fp[-12] - fp[4]) * (2.0f * fp[-4] - fp[4] - fp[-12]) +
                (fp[-11] - fp[5]) * (2.0f * fp[-3] - fp[5] - fp[-11])) /
                    (2.0f * h1) - 1;
        f2=    ((fp[-4] - fp[12]) * (2.0f * fp[4] - fp[12] - fp[-4]) +
                (fp[-3] - fp[13]) * (2.0f * fp[5] - fp[13] - fp[-3])) /
                    (2.0f * h2) + 1;

            /* get sample mean and variance of the three */
        m = 0.333333f * (pfreq + f1 + f2);
        var = 0.5f * ((pfreq-m)*(pfreq-m) + (f1-m)*(f1-m) + (f2-m)*(f2-m));

        totalfreq = i + m;
        if (var * total_power > KNOCKTHRESH * height || var < 1e-30)
        {
#if 0
            if (x->x_nprint)
                post("cancel: %.2f hz, index %.1f, power %.5f, stdev=%.2f",
                    totalfreq * hzperbin, BPERO_OVER_LOG2 * log(totalfreq) - 96,
                     height, sqrt(var));
#endif
            continue;
        }
        stdev = fsqrt(var);
        if (totalfreq < 4)
        {
            if (x->x_nprint) post("oops: was %d,  freq %f, m %f, stdev %f h %f",
                i,  totalfreq, m, stdev, height);
            totalfreq = 4;
        }
        pk1->p_width = stdev;

        pk1->p_pow = height;
        pk1->p_loudness = fsqrt(fsqrt(height));
        pk1->p_fp = fp;
        pk1->p_freq = totalfreq;
        npeak++;
#if 1
        if (x->x_nprint)
        {
            post("peak: %.2f hz. index %.1f, power %.5f, stdev=%.2f",
                pk1->p_freq * hzperbin,
                BPERO_OVER_LOG2 * log(pk1->p_freq) - 96,
                 height, stdev);
        }
#endif
        pk1++;
    }

            /* prepare the raw peaks for output */
    for (i = 0, pk1 = peaklist, pk2 = x->x_peakbuf; i < npeak;
        i++, pk1++, pk2++)
    {
        double loudness = pk1->p_loudness;
        if (i >= npeakout) break;
        pk2->po_freq = hzperbin * pk1->p_freq;
        pk2->po_amp = (2.f / (double)n) * (loudness * loudness);
    }
    for (; i < npeakout; i++, pk2++) pk2->po_amp = pk2->po_freq = 0;

        /* now, working back into spect2, make a sort of "liklihood"
         * spectrum.  Proceeding in 48ths of an octave,  from 2 to
         * n/2 (in bins), the likelihood of each pitch range is contributed
         * to by every peak in peaklist that's an integer multiple of it
         * in frequency.
         */

    if (npeak > npeakanal) npeak = npeakanal; /* max # peaks to analyze */
    {double*fp1;
    for (i = 0, fp1 = histogram; i < maxbin; i++) *fp1++ = 0;
    }
    for (i = 0, pk1 = peaklist; i < npeak; i++, pk1++)
    {
        double pit = BPERO_OVER_LOG2 * flog(pk1->p_freq) - 96.0f;
        double binbandwidth = FACTORTOBINS * pk1->p_width/pk1->p_freq;
        double putbandwidth = (binbandwidth < 2 ? 2 : binbandwidth);
        double weightbandwidth = (binbandwidth < 1.0f ? 1.0f : binbandwidth);
        /* double weightamp = 1.0f + 3.0f * pk1->p_pow / pow; */
        double weightamp = 4. * pk1->p_loudness / total_loudness;
        const double* fp4;
        for (j = 0, fp4 = sigfiddle_partialonset; j < NPARTIALONSET; j++, fp4++)
        {
            double bin = pit - *fp4;
            if (bin < maxbin)
            {
                double* fp3;
                double para, pphase, score = 30.0f * weightamp /
                    ((j+x->x_npartial) * weightbandwidth);
                int firstbin = bin + 0.5f - 0.5f * putbandwidth;
                int lastbin = bin + 0.5f + 0.5f * putbandwidth;
                int ibw = lastbin - firstbin;
                if (firstbin < -BINGUARD) break;
                para = 1.0f / (putbandwidth * putbandwidth);
                
                
                
                for (k = 0, fp3 = histogram + firstbin,
                    pphase = firstbin-bin; k <= ibw;
                        k++, fp3++,  pphase += 1.0f)
                {
                    *fp3 += score * (1.0f - para * pphase * pphase);
                }
            }
        }
    }
#if 1
    if (x->x_nprint)
    {
        for (i = 0; i < 6*5; i++)
        {
            double fhz = hzperbin * exp ((8*i + 96) * (1./BPERO_OVER_LOG2));
            if (!(i % 6)) post("-- bin %d pitch %f freq %f----", 8*i,
                ftom(fhz), fhz);;
            post("%3d %3d %3d %3d %3d %3d %3d %3d",
                (int)(histogram[8*i]),
                (int)(histogram[8*i+1]),
                (int)(histogram[8*i+2]),
                (int)(histogram[8*i+3]),
                (int)(histogram[8*i+4]),
                (int)(histogram[8*i+5]),
                (int)(histogram[8*i+6]),
                (int)(histogram[8*i+7]));
        }
    }

#endif

        /*
         * Next we find up to NPITCH strongest peaks in the histogram.
         * if a peak is related to a stronger one via an interval in
         * the sigfiddle_partialonset array,  we suppress it.
         */

    for (npitch = 0; npitch < x->x_npitch; npitch++)
    {
        int indx;
        double best;
        if (npitch)
        {
            for (best = 0, indx = -1, j=1; j < maxbin-1; j++)
            {
                if (histogram[j] > best && histogram[j] > histogram[j-1] &&
                    histogram[j] > histogram[j+1])
                {
                    for (k = 0; k < npitch; k++)
                        if (histvec[k].h_index == j)
                            goto peaknogood;
                    for (k = 0; k < NPARTIALONSET; k++)
                    {
                        if (j - sigfiddle_intpartialonset[k] < 0) break;
                        if (histogram[j - sigfiddle_intpartialonset[k]]
                            > histogram[j]) goto peaknogood;
                    }
                    for (k = 0; k < NPARTIALONSET; k++)
                    {
                        if (j + sigfiddle_intpartialonset[k] >= maxbin) break;
                        if (histogram[j + sigfiddle_intpartialonset[k]]
                            > histogram[j]) goto peaknogood;
                    }
                    indx = j;
                    best = histogram[j];
                }
            peaknogood: ;
            }
        }
        else
        {
            for (best = 0, indx = -1, j=0; j < maxbin; j++)
                if (histogram[j] > best)
                    indx = j,  best = histogram[j];
        }
        if (indx < 0) break;
        histvec[npitch].h_value = best;
        histvec[npitch].h_index = indx;
    }
#if 1
    if (x->x_nprint)
    {
        for (i = 0; i < npitch; i++)
        {
            post("index %d freq %f --> value %f", histvec[i].h_index,
                exp((1./BPERO_OVER_LOG2) * (histvec[i].h_index + 96)),
                histvec[i].h_value);
            post("next %f , prev %f",
                exp((1./BPERO_OVER_LOG2) * (histvec[i].h_index + 97)),
                exp((1./BPERO_OVER_LOG2) * (histvec[i].h_index + 95)) );
        }
    }
#endif

        /* for each histogram peak, we now search back through the
         * FFT peaks.  A peak is a pitch if either there are several
         * harmonics that match it,  or else if (a) the fundamental is
         * present,  and (b) the sum of the powers of the contributing peaks
         * is at least 1/100 of the total power.
         *
         * A peak is a contributor if its frequency is within 25 cents of
         * a partial from 1 to 16.
         *
         * Finally, we have to be at least 5 bins in frequency, which
         * corresponds to 2-1/5 periods fitting in the analysis window.
         */

    for (i = 0; i < npitch; i++)
    {
        double cumpow = 0, cumstrength = 0, freqnum = 0, freqden = 0;
        int npartials = 0,  nbelow8 = 0;
            /* guessed-at frequency in bins */
        double putfreq = fexp((1.0f / BPERO_OVER_LOG2) *
            (histvec[i].h_index + 96.0f));
        for (j = 0; j < npeak; j++)
        {
            double fpnum = peaklist[j].p_freq/putfreq;
            int pnum = fpnum + 0.5f;
            double fipnum = pnum;
            double deviation;
            if (pnum > 16 || pnum < 1) continue;
            deviation = 1.0f - fpnum/fipnum;
            if (deviation > -PARTIALDEVIANCE && deviation < PARTIALDEVIANCE)
            {
                /*
                 * we figure this is a partial since it's within 1/4 of
                 * a halftone of a multiple of the putative frequency.
                 */

                double stdev, weight;
                npartials++;
                if (pnum < 8) nbelow8++;
                cumpow += peaklist[j].p_pow;
                cumstrength += fsqrt(fsqrt(peaklist[j].p_pow));
                stdev = (peaklist[j].p_width > MINBW ?
                    peaklist[j].p_width : MINBW);
                weight = 1.0f / ((stdev*fipnum) * (stdev*fipnum));
                freqden += weight;
                freqnum += weight * peaklist[j].p_freq/fipnum;          
#if 1
                if (x->x_nprint)
                {
                    post("peak %d partial %d f=%f w=%f",
                        j, pnum, peaklist[j].p_freq/fipnum, weight);
                }
#endif
            }
#if 1
            else if (x->x_nprint) post("peak %d partial %d dev %f",
                        j, pnum, deviation);
#endif
        }
        if ((nbelow8 < 4 || npartials < 7) && cumpow < 0.01f * total_power)
            histvec[i].h_value = 0;
        else
        {
            double pitchpow = (cumstrength * cumstrength) *
                (cumstrength * cumstrength);
            double freqinbins = freqnum/freqden;
                /* check for minimum output frequency */

            if (freqinbins < MINFREQINBINS)
                histvec[i].h_value = 0;
            else
            {
                    /* we passed all tests... save the values we got */
                histvec[i].h_pitch = ftom(hzperbin * freqnum/freqden);
                histvec[i].h_loud = (100.0f -DBFUDGE) +
                    (LOGTODB) * log(pitchpow/n);
            }
        }
    }
#if 1
    if (x->x_nprint)
    {
        for (i = 0; i < npitch; i++)
        {
            if (histvec[i].h_value > 0)
                post("index %d pit %f loud %f", histvec[i].h_index,
                histvec[i].h_pitch, histvec[i].h_loud);
            else post("-- cancelled --");
        }
    }
#endif

        /* now try to find continuous pitch tracks that match the new
         * pitches.  First mark each peak unmatched.
         */
    for (i = 0, hp1 = histvec; i < npitch; i++, hp1++)
        hp1->h_used = 0;

        /* for each old pitch, try to match a new one to it. */
    for (i = 0, phist = x->x_hist; i < x->x_npitch; i++,  phist++)
    {
        double thispitch = phist->h_pitches[oldphase];
        phist->h_pitch = 0;         /* no output, thanks */
        phist->h_wherefrom = 0;
        if (thispitch == 0.0f) continue;
        for (j = 0, hp1 = histvec; j < npitch; j++, hp1++)
            if ((hp1->h_value > 0) && hp1->h_pitch > thispitch - GLISS
                && hp1->h_pitch < thispitch + GLISS)
        {
            phist->h_wherefrom = hp1;
            hp1->h_used = 1;
        }
    }
    for (i = 0, hp1 = histvec; i < npitch; i++, hp1++)
        if ((hp1->h_value > 0) && !hp1->h_used)
    {
        for (j = 0, phist = x->x_hist; j < x->x_npitch; j++,  phist++)
            if (!phist->h_wherefrom)
        {
            phist->h_wherefrom = hp1;
            phist->h_age = 0;
            phist->h_noted = 0;
            hp1->h_used = 1;
            goto happy;
        }
        break;
    happy: ;
    }
        /* copy the pitch info into the history vector */
    for (i = 0, phist = x->x_hist; i < x->x_npitch; i++,  phist++)
    {
        if (phist->h_wherefrom)
        {
            phist->h_amps[newphase] = phist->h_wherefrom->h_loud;
            phist->h_pitches[newphase] =
                phist->h_wherefrom->h_pitch;
            (phist->h_age)++;
        }
        else
        {
            phist->h_age = 0;
            phist->h_amps[newphase] = phist->h_pitches[newphase] = 0;
        }
    }
#if 1
    if (x->x_nprint)
    {
        post("vibrato %d %f", x->x_vibbins, x->x_vibdepth);
        for (i = 0, phist = x->x_hist; i < x->x_npitch; i++,  phist++)
        {
            post("noted %f, age %d", phist->h_noted,  phist->h_age);
#ifndef I860
            post("values %f %f %f %f %f",
                phist->h_pitches[newphase],
                phist->h_pitches[(newphase + HISTORY-1)%HISTORY],
                phist->h_pitches[(newphase + HISTORY-2)%HISTORY],
                phist->h_pitches[(newphase + HISTORY-3)%HISTORY],
                phist->h_pitches[(newphase + HISTORY-4)%HISTORY]);
#endif
        }
    }
#endif
        /* look for envelope attacks */

    x->x_attackvalue = 0;

    if (x->x_peaked)
    {
        if (total_db > x->x_amphi)
        {
            int binlook = newphase - x->x_attackbins;
            if (binlook < 0) binlook += HISTORY;
            if (total_db > x->x_dbs[binlook] + x->x_attackthresh)
            {
                x->x_attackvalue = 1;
                x->x_peaked = 0;
            }
        }
    }
    else
    {
        int binlook = newphase - x->x_attackbins;
        if (binlook < 0) binlook += HISTORY;
        if (x->x_dbs[binlook] > x->x_amphi && x->x_dbs[binlook] > total_db)
            x->x_peaked = 1;
    }

        /* for each current frequency track, test for a new note using a
         * stability criterion.  Later perhaps we should also do as in
         * pitch~ and check for unstable notes a posteriori when
         * there's a new attack with no note found since the last onset;
         * but what's an attack &/or onset when we're polyphonic?
         */

    for (i = 0, phist = x->x_hist; i < x->x_npitch; i++,  phist++)
    {
            /*
             * if we've found a pitch but we've now strayed from it turn
             * it off.
             */
        if (phist->h_noted)
        {
            if (phist->h_pitches[newphase] > phist->h_noted + x->x_vibdepth
                || phist->h_pitches[newphase] < phist->h_noted - x->x_vibdepth)
                    phist->h_noted = 0;
        }
        else
        {
            if (phist->h_wherefrom && phist->h_age >= x->x_vibbins)
            {
                double centroid = 0;
                int notvar = 0;
                for (j = 0, k = newphase; j < x->x_vibbins; j++)
                {
                    centroid += phist->h_pitches[k];
                    k--;
                    if (k < 0) k = HISTORY-1;
                }
                centroid /= x->x_vibbins;
                for (j = 0, k = newphase; j < x->x_vibbins; j++)
                {
                        /* calculate deviation from norm */
                    double dev = centroid - phist->h_pitches[k];
                    k--;
                    if (k < 0) k = HISTORY-1;
                    if (dev > x->x_vibdepth || -dev > x->x_vibdepth) notvar = 1;
                }
                if (!notvar)
                {
                    phist->h_pitch = phist->h_noted = centroid;
                }
            }
        }
    }
    return;

nopow:
    for (i = 0; i < x->x_npitch; i++)
    {
        x->x_hist[i].h_pitch = x->x_hist[i].h_noted =
            x->x_hist[i].h_pitches[newphase] =
            x->x_hist[i].h_amps[newphase] = 0;
        x->x_hist[i].h_age = 0;
    }
    x->x_peaked = 1;
    x->x_dbage = 0;
}

void sigfiddle_auto(t_sigfiddle *x, const double f)
{
    x->x_auto = (f != 0);
}

static void sigfiddle_freebird(t_sigfiddle *x)
{
    if (x->x_inbuf)
    {
        delete[] x->x_inbuf;
        x->x_inbuf = 0;
    }
    if (x->x_lastanalysis)
    {
        delete[] x->x_lastanalysis;
        x->x_lastanalysis = 0;
    }
    if (x->x_spiral)
    {
        delete[] x->x_spiral;
        x->x_spiral = 0;
    }
    x->x_hop = 0;
}

void sigfiddle_setnpoints(t_sigfiddle *x, const double fnpoints)
{
    int i, npoints = fnpoints;
    sigfiddle_freebird(x);
    if (npoints < MINPOINTS || npoints > MAXPOINTS)
    {
        error("fiddle~: npoints out of range; using %d",
            npoints = DEFAULTPOINTS);
    }
    if (npoints != (1 << sigfiddle_ilog2(npoints)))
    {
        error("fiddle~: npoints not a power of 2; using %d", 
            npoints = (1 << sigfiddle_ilog2(npoints)));
    }
    x->x_hop = npoints >> 1;
    x->x_inbuf = new double[x->x_hop];
    x->x_lastanalysis = new double[2 * x->x_hop + 4 * FILTSIZE];
    x->x_spiral = new double[2 * x->x_hop];
    for (i = 0; i < x->x_hop; i++)
        x->x_inbuf[i] = 0;
    for (i = 0; i < npoints + 4 * FILTSIZE; i++)
        x->x_lastanalysis[i] = 0;
    for (i = 0; i < x->x_hop; i++)
        x->x_spiral[2*i] =    cos((3.14159*i)/(npoints)),
        x->x_spiral[2*i+1] = -sin((3.14159*i)/(npoints));
    x->x_phase = 0;
}

int sigfiddle_doinit(t_sigfiddle *x, long npoints, long npitch,
    long npeakanal, long npeakout)
{
    t_peakout *buf4;
    int i;

    if (!npeakanal && !npeakout) npeakanal = DEFNPEAK, npeakout = 0;
    if (!npeakanal < 0) npeakanal = 0;
    else if (npeakanal > MAXPEAK) npeakanal = MAXPEAK;
    if (!npeakout < 0) npeakout = 0;
    else if (npeakout > MAXPEAK) npeakout = MAXPEAK;
    if (npitch <= 0) npitch = 0;
    else if (npitch > MAXNPITCH) npitch = MAXNPITCH;
    if (npeakanal && !npitch) npitch = 1;
    if (!npoints)
        npoints = DEFAULTPOINTS;

    sigfiddle_setnpoints(x, npoints);

    buf4 = new t_peakout[npeakout]; /* FIXME: memory leak */
    for (i = 0; i < npeakout; i++)
        buf4[i].po_freq = buf4[i].po_amp = 0;
    x->x_peakbuf = buf4;

    x->x_npeakout = npeakout;
    x->x_npeakanal = npeakanal;
    x->x_phase = 0;
    x->x_histphase = 0;
    x->x_sr = 44100;            /* this and the next are filled in later */
    for (i = 0; i < MAXNPITCH; i++)
    {
        int j;
        x->x_hist[i].h_pitch = x->x_hist[i].h_noted = 0;
        x->x_hist[i].h_age = 0;
        x->x_hist[i].h_wherefrom = 0;
        for (j = 0; j < HISTORY; j++)
            x->x_hist[i].h_amps[j] = x->x_hist[i].h_pitches[j] = 0;
    }
    x->x_nprint = 0;
    x->x_npitch = npitch;
    for (i = 0; i < HISTORY; i++) x->x_dbs[i] = 0;
    x->x_dbage = 0;
    x->x_peaked = 0;
    x->x_auto = 1;
    x->x_amplo = DEFAMPLO;
    x->x_amphi = DEFAMPHI;
    x->x_attacktime = DEFATTACKTIME;
    x->x_attackbins = 1;                /* real value calculated afterward */
    x->x_attackthresh = DEFATTACKTHRESH;
    x->x_vibtime = DEFVIBTIME;
    x->x_vibbins = 1;                   /* real value calculated afterward */
    x->x_vibdepth = DEFVIBDEPTH;
    x->x_npartial = 7;
    x->x_attackvalue = 0;
    return (1);
}


#if 0 /* MAX26-SPECIFIC */
    t_sig *x_io[IN1+OUT0];          /* number of signal inputs and outputs */
#endif

class sigfiddle_class
{
public:
    /* OUTLETS: */
    double x_envout;
    bool x_attackout; // type: BANG
    double x_noteout;
    struct { double pitch, amp; } h_outlet[MAXNPITCH];

private:
    sigfiddle_class(const sigfiddle_class&);
    void operator =(const sigfiddle_class&);
public:
    sigfiddle_class(long npoints, long npitch, long npeakanal, long npeakout)
    {
        memset(&x, 0, sizeof(x));
        /*
        x.x_inbuf        = 0;
        x.x_lastanalysis = 0;
        x.x_spiral       = 0;
        */
        
        sigfiddle_doinit(&x, npoints, npitch, npeakanal, npeakout);
        x_attackout = false;
        x_envout   = 0;
        x_noteout  = 0;
        for(unsigned n=0; n<MAXNPITCH; ++n) h_outlet[n].pitch = h_outlet[n].amp = 0;
    }
    
    ~sigfiddle_class()
    {
        sigfiddle_freebird(&x);
    }
    
    void begin_dsp(double sample_rate)
    {
        //u_stdout(x);
        
        x.x_sr = sample_rate;
        
        this->reattack(x.x_attacktime, x.x_attackthresh);
        this->vibrato(x.x_vibtime, x.x_vibdepth);
    }
    
    void put_dsp(double *in1, int n)
    {
        int count;
        double *fp;
        for (count = 0, fp = x.x_inbuf + x.x_phase;
                count < n; count++) *fp++ = *in1++;
         
        //post("hop = %d\n", x.x_hop);
        
        if (fp == x.x_inbuf + x.x_hop)
        {
            sigfiddle_doit(&x);
            x.x_phase = 0;
            if (x.x_auto) this->report_output();
            if (x.x_nprint) x.x_nprint--;
        }
        else x.x_phase += n;
    }

    void debug() { x.x_nprint = 1; }
    
    void print()
    {
        post("npoints %d,",  2 * x.x_hop);
        post("amp-range %f %f,",  x.x_amplo, x.x_amphi);
        post("reattack %d %f,",  x.x_attacktime, x.x_attackthresh);
        post("vibrato %d %f",  x.x_vibtime, x.x_vibdepth);
        post("npartial %f",  x.x_npartial);
        post("auto %d",  x.x_auto);
    }

    void amprange(double amplo, double amphi)
    {
        if (amplo < 0) amplo = 0;
        if (amphi < amplo) amphi = amplo + 1;
        x.x_amplo = amplo;
        x.x_amphi = amphi;
    }

    void reattack(double attacktime, double attackthresh)
    {
        if (attacktime < 0) attacktime = 0;
        if (attackthresh <= 0) attackthresh = 1000;
        x.x_attacktime = attacktime;
        x.x_attackthresh = attackthresh;
        x.x_attackbins = (x.x_sr * 0.001 * attacktime) / x.x_hop;
        if (x.x_attackbins >= HISTORY) x.x_attackbins = HISTORY - 1;
    }

    void vibrato(long vibtime, double vibdepth)
    {
        if (vibtime < 0) vibtime = 0;
        if (vibdepth <= 0) vibdepth = 1000;
        x.x_vibtime = vibtime;
        x.x_vibdepth = vibdepth;
        x.x_vibbins = (x.x_sr * 0.001  * vibtime) / x.x_hop;
        if (x.x_vibbins >= HISTORY) x.x_vibbins = HISTORY - 1;
        if (x.x_vibbins < 1) x.x_vibbins = 1;
    }

    void npartial(double npartial)
    {
        if (npartial < 0.1) npartial = 0.1;
        x.x_npartial = npartial;
    }
    
    void dump()
    {
      post(
          "Got: envout=%g, noteout=%g",
              this->x_envout,
              this->x_noteout);
      for(unsigned a=0; a<MAXNPITCH; ++a)
        if(this->h_outlet[a].pitch != 0.0)
          post("%u: pit=%g amp=%g",
              a, this->h_outlet[a].pitch,
                 this->h_outlet[a].amp);
    }

    void report_output()
    {
        int i;
        t_pitchhist *ph;
        
        x_envout = x.x_dbs[x.x_histphase];
        
        for (i = 0,  ph = x.x_hist; i < x.x_npitch; i++,  ph++)
        {
            h_outlet[i].pitch = ph->h_pitches[x.x_histphase];
            h_outlet[i].amp   = ph->h_amps[x.x_histphase];
        }
        
        x_attackout = x.x_attackvalue;
        
        for (i = 0,  ph = x.x_hist; i < x.x_npitch; i++,  ph++)
            if (ph->h_pitch)
                x_noteout = ph->h_pitch;
        
        //printf("Reported output: "); dump();
    }

private:
    t_sigfiddle x;
};
