#include <iostream>

#include <map>
#include <vector>
#include <stdio.h>
#include <string>
#include "ch.hh"

static std::string filename;
static unsigned    trackno;

class NumStat
{
    unsigned n;
    double min,max,sum;
    double sqr_sum;
public:
    NumStat(): n(0),min(0),max(0),sum(0) { }
    void Cumulate(double val)
    {
        sum+=val; sqr_sum += val*val;
        if(!n||val<min)min=val;
        if(!n||val>max)max=val;
        ++n;
    }
    double GetAverage() const { return sum/n; }
    double GetMin() const { return min; }
    double GetMax() const { return max; }
    unsigned GetCount() const { return n; }
    double GetVariance() const { return (sqr_sum - sum*GetAverage()) / (n-1); }
};

typedef std::map<unsigned/*chn*/,int/*tone*/> chstatus_t;
typedef std::map<unsigned/*time*/, chstatus_t> melody_t;

static const std::string NameTone(int tone)
{
    static const char* const tonenames[12] =
    {
        "C-","C#","D-","D#","E-","F-","F#","G-","G#","A-","A#","B-"
    };
    char result[16];
    sprintf(result, "%s%d", tonenames[(tone+3)%12], (tone+3)/12-6);
    return result;
}

static const std::string sqlfix(const std::string &q)
{
    const char *s = q.c_str();
    std::string diu = "'";
    while(*s)
    {        
        if(*s=='\'')diu += "\\'";
        else diu += *s;
        ++s;
    }       
    return diu + "'";
}

static void SaveMelody(const melody_t& track, unsigned loop_begin)
{
    const double fps = 60.0;
    
    std::map<int, std::vector<int> > channels;
    std::map<int, std::vector<unsigned> > timings;
    
    {std::map<int, int> prevtones;
    for(melody_t::const_iterator i = track.begin(); i != track.end(); ++i)
    {
        unsigned timepos = i->first;
        for(chstatus_t::const_iterator j = i->second.begin(); j != i->second.end(); ++j)
        {
            int chno = j->first;
            int tonediff = j->second - prevtones[chno];
            prevtones[chno] = j->second;
            channels[chno].push_back(tonediff);
            timings[chno].push_back(timepos);
        }
    }}
    
    std::cout<< "insert into song set filename=" << sqlfix(filename)
             << ",trackno=" << trackno
             << ",loopbegin=" << (loop_begin / fps)
             << ";\n";

    std::cout<< "set @songid = last_insert_id();\n";

    for(std::map<int, std::vector<int> >::const_iterator
        i = channels.begin(); i != channels.end(); ++i)
    {
        int chno = i->first;
        const std::vector<int>& trk = i->second;
        std::string data;
        for(unsigned a=0; a<trk.size(); ++a)
            data += (char)(128 + trk[a]);
        
        std::cout<< "insert into song_channel values(@songid, " << chno << "," << sqlfix(data) << ");\n";
        
        for(unsigned b=0; b<trk.size(); ++b)
        {
            std::cout<< "insert into song_index values(@songid, " << chno
                     << "," << b
                     << "," << (timings[chno][b] / fps)
                     << "," << sqlfix(data.substr(b, 32)) << ");\n";
        }
    }
}

static void SaveMelody(const std::vector<situation>& track, unsigned loop_begin)
{
    std::map<unsigned, NumStat> melodic_deviation;
    unsigned max_channel=0;
    
    const double standard_tuning = fmod(log2(440)*12.0, 1.0);
    
    for(unsigned a=0; a<track.size(); ++a)
        for(situation::const_iterator
            i = track[a].begin();
            i != track[a].end();
            ++i)
        {
            if(i->first > max_channel) max_channel = i->first;
            double tone = log2(i->second.hz)*12.0;
            tone -= standard_tuning;
            double difference = fmod(tone, 1.0);
            if(difference > 0.5) difference -= 1.0;
            melodic_deviation[i->first].Cumulate(difference);
        }
    for(std::map<unsigned, NumStat>::const_iterator
        j = melodic_deviation.begin();
        j != melodic_deviation.end();
        ++j)
    {
        fprintf(stderr,
            "Trk %u: deviation %lf semitones (%lf..%lf), variance %lf from standard tuning (%u samples)\n",
            j->first,
            j->second.GetAverage(),  j->second.GetMin(), j->second.GetMax(),
            j->second.GetVariance(),
            j->second.GetCount()
        );
    }
    
    std::vector<double> prev_tone(max_channel+1);
    
    // tone 105 = 440 Hz "A". tone 106 = "B", tone 107 = "C" etc.
    
    melody_t melody_simplified;
    for(unsigned a=0; a<track.size(); ++a)
    {
        for(situation::const_iterator
            i = track[a].begin();
            i != track[a].end();
            ++i)
        {
            double tone = (log2(i->second.hz)*12.0 - standard_tuning)
                        - melodic_deviation[i->first].GetAverage();
            const double abs_tonediff = fabs(prev_tone[i->first] - tone);
            prev_tone[i->first] = tone;
            
            double prev_volume = 0.0;
            if(a > 0) { situation::const_iterator j = track[a-1].find(i->first);
                        if(j != track[a-1].end()) prev_volume = j->second.amplitude; }
            
            double volume = i->second.amplitude;
            
            if(prev_volume < silence_level
            || volume >= prev_volume + 2/16.0
            || (prev_volume >= silence_level && volume >= prev_volume*3)
            || abs_tonediff >= 0.92856
            || a == loop_begin)
            {
                // It's a tick point.
                melody_simplified[a][i->first] = (int)(tone+0.5);
            }
        }
    }
    
    SaveMelody(melody_simplified, loop_begin);
}

int main(int argc, char**argv)
{
    filename = argv[1];
    trackno  = std::atoi(argv[2]);

    unsigned track_size, loop_begin;
    scanf("%u %u", &track_size, &loop_begin);
    std::vector<situation> track(track_size);
    for(unsigned a=0; a<track.size(); ++a)
    {
        unsigned num_chs; scanf("%u", &num_chs);
        situation line;
        for(unsigned n=0; n<num_chs; ++n)
        {
            unsigned chno;
            ch ch;
            scanf("%u %lf %lf", &chno, &ch.amplitude, &ch.hz);
            line[chno] = ch;
        }
        track[a] = line;
    }
    SaveMelody(track, loop_begin);
}
